"""Test the Abode device classes."""
import jaraco.abode
from jaraco.abode.helpers import urls
import pytest

from .mock import login as LOGIN
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import logout as LOGOUT
from .mock import panel as PANEL
from .mock import devices as DEVICES
from .mock.devices import alarm as ALARM


class TestAlarm:
    """Test the Alarm class."""

    def test_abode_alarm_setup(self, m):
        """Check that Abode alarm device is set up properly."""
        panel = PANEL.get_response_ok(mode='standby')
        alarm = ALARM.device(area='1', panel=panel)

        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        alarm_device = self.client.get_alarm()

        assert alarm_device is not None

        assert alarm_device._state == alarm

    def test_alarm_device_properties(self, m):
        """Check that the abode device properties are working."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(
            urls.PANEL,
            json=PANEL.get_response_ok(
                mode='standby',
                battery=True,
                is_cellular=True,
                mac='01:AA:b3:C4:d5:66',
            ),
        )
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)

        # Logout to reset everything
        self.client.logout()

        # Get alarm and test
        alarm = self.client.get_alarm()
        assert alarm.id == 'area_1'
        assert alarm is not None
        assert alarm.mode == 'standby'
        assert alarm.status == 'standby'
        assert alarm.battery
        assert alarm.is_cellular
        assert not alarm.is_on
        assert alarm.uuid == '01aab3c4d566'
        assert alarm.mac_address == '01:AA:b3:C4:d5:66'

        # Change alarm properties and state to away and test
        m.get(
            urls.PANEL,
            json=PANEL.get_response_ok(mode='away', battery=False, is_cellular=False),
        )

        # Refresh alarm and test
        alarm.refresh()

        assert alarm.id == 'area_1'
        assert alarm.mode == 'away'
        assert alarm.status == 'away'
        assert not alarm.battery
        assert not alarm.is_cellular
        assert alarm.is_on

        # Change alarm state to final on state and test
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='home'))

        # Refresh alarm and test
        alarm.refresh()
        assert alarm.mode == 'home'
        assert alarm.status == 'home'
        assert alarm.is_on

    def test_alarm_device_mode_changes(self, m):
        """Test that the abode alarm can change/report modes."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)

        # Logout to reset everything
        self.client.logout()

        # Assert that after login we have our alarm device with standby mode
        alarm = self.client.get_alarm()

        assert alarm is not None
        assert alarm.status == 'standby'

        # Set mode URLs
        m.put(
            urls.panel_mode('1', 'standby'),
            json=PANEL.put_response_ok(mode='standby'),
        )
        m.put(
            urls.panel_mode('1', 'away'),
            json=PANEL.put_response_ok(mode='away'),
        )
        m.put(
            urls.panel_mode('1', 'home'),
            json=PANEL.put_response_ok(mode='home'),
        )

        # Set and test text based mode changes
        assert alarm.set_mode('home')
        assert alarm.mode == 'home'
        assert not alarm.is_standby
        assert alarm.is_home
        assert not alarm.is_away

        assert alarm.set_mode('away')
        assert alarm.mode == 'away'
        assert not alarm.is_standby
        assert not alarm.is_home
        assert alarm.is_away

        assert alarm.set_mode('standby')
        assert alarm.mode == 'standby'
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        # Set and test direct mode changes
        assert alarm.set_home()
        assert alarm.mode == 'home'
        assert not alarm.is_standby
        assert alarm.is_home
        assert not alarm.is_away

        assert alarm.set_away()
        assert alarm.mode == 'away'
        assert not alarm.is_standby
        assert not alarm.is_home
        assert alarm.is_away

        assert alarm.set_standby()
        assert alarm.mode == 'standby'
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        # Set and test default mode changes
        assert alarm.switch_off()
        assert alarm.mode == 'standby'
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        self.client.set_default_mode('home')
        assert alarm.switch_on()
        assert alarm.mode == 'home'
        assert not alarm.is_standby
        assert alarm.is_home
        assert not alarm.is_away

        assert alarm.switch_off()
        assert alarm.mode == 'standby'
        assert alarm.is_standby
        assert not alarm.is_home
        assert not alarm.is_away

        self.client.set_default_mode('away')
        assert alarm.switch_on()
        assert alarm.mode == 'away'
        assert not alarm.is_standby
        assert not alarm.is_home
        assert alarm.is_away

        # Test that no mode throws exception
        with pytest.raises(jaraco.abode.Exception):
            alarm.set_mode(mode=None)

        # Test that an invalid mode throws exception
        with pytest.raises(jaraco.abode.Exception):
            alarm.set_mode('chestnuts')

        # Test that an invalid mode change response throws exception
        m.put(
            urls.panel_mode('1', 'home'),
            json=PANEL.put_response_ok(mode='away'),
        )

        with pytest.raises(jaraco.abode.Exception):
            alarm.set_mode('home')

        # Test that an invalid area in mode change response throws exception
        m.put(
            urls.panel_mode('1', 'home'),
            json=PANEL.put_response_ok(area='2', mode='home'),
        )

        with pytest.raises(jaraco.abode.Exception):
            alarm.set_mode('home')



import logging
import warnings
from typing import Tuple

from jaraco.classes.ancestry import iter_subclasses

import jaraco
from ..helpers import errors as ERROR
from ..helpers import urls
from ..state import Stateful
from .control import needs_control_url
from . import pkg


log = logging.getLogger(__name__)


class Device(Stateful):
    """Class to represent each Abode device."""

    tags: Tuple[str, ...] = ()
    _desc_t = '{name} (ID: {id}, UUID: {uuid}) - {type} - {status}'
    _url_t = urls.DEVICE

    @needs_control_url
    def set_status(self, status):
        """Set device status."""
        path = self._state['control_url']

        status_data = {'status': str(status)}

        response = self._client.send_request(method="put", path=path, data=status_data)
        response_object = response.json()

        log.debug("Set Status Response: %s", response.text)

        if response_object['id'] != self.id:
            raise jaraco.abode.Exception(ERROR.SET_STATUS_DEV_ID)

        if response_object['status'] != str(status):
            raise jaraco.abode.Exception(ERROR.SET_STATUS_STATE)

        # Note: Status result is of int type, not of new status of device.
        # Seriously, why would you do that?
        # So, can't set status here must be done at device level.

        log.info("Set device %s status to: %s", self.id, status)

    @needs_control_url
    def set_level(self, level):
        """Set device level."""
        url = self._state['control_url']

        level_data = {'level': str(level)}

        response = self._client.send_request("put", url, data=level_data)
        response_object = response.json()

        log.debug("Set Level Response: %s", response.text)

        if response_object['id'] != self.id:
            raise jaraco.abode.Exception(ERROR.SET_STATUS_DEV_ID)

        if response_object['level'] != str(level):
            raise jaraco.abode.Exception(ERROR.SET_STATUS_STATE)

        self.update(response_object)

        log.info("Set device %s level to: %s", self.id, level)

    def get_value(self, name):
        """Get a value from the device state."""
        return self._state.get(name.lower(), {})

    @property
    def status(self):
        """Shortcut to get the generic status of a device."""
        return self.get_value('status')

    @property
    def battery_low(self):
        """Is battery level low."""
        return bool(int(self.get_value('faults').get('low_battery', '0')))

    @property
    def no_response(self):
        """Is the device responding."""
        return bool(int(self.get_value('faults').get('no_response', '0')))

    @property
    def out_of_order(self):
        """Is the device out of order."""
        return bool(int(self.get_value('faults').get('out_of_order', '0')))

    @property
    def tampered(self):
        """Has the device been tampered with."""
        # 'tempered' - Typo in API?
        return bool(int(self.get_value('faults').get('tempered', '0')))

    @property
    def name(self):
        """Get the name of this device."""
        fallback = f'{self.type} {self.id}'
        return self._state.get('name') or fallback

    @property
    def device_id(self):
        """Get the device id."""
        warnings.warn("Device.device_id is deprecated. Use .id.", DeprecationWarning)
        return self.id

    @property
    def device_uuid(self):
        """Get the device uuid."""
        warnings.warn(
            "Device.device_uuid is deprecated. Use .uuid.", DeprecationWarning
        )
        return self.uuid

    @staticmethod
    def resolve_type_unknown(state):
        if state['generic_type'] != 'unknown':
            return

        from .sensor import Sensor

        if Sensor.is_sensor(state):
            state['generic_type'] = 'sensor'
            return

        version = state.get('version', '')

        state['generic_type'] = (
            'occupancy' if version.startswith('MINIPIR') else 'motion'
        )

    @classmethod
    def new(cls, state, client):
        """Create new device object for the given type."""
        pkg.import_all()
        try:
            type_tag = state['type_tag']
        except KeyError as exc:
            raise jaraco.abode.Exception(ERROR.UNABLE_TO_MAP_DEVICE) from exc

        state['generic_type'] = cls.get_generic_type(type_tag)
        cls.resolve_type_unknown(state)
        sensor = cls.by_type().get(state['generic_type'], lambda *args: None)
        return sensor(state, client)

    @classmethod
    def by_type(cls):
        return {sub_cls.__name__.lower(): sub_cls for sub_cls in iter_subclasses(cls)}

    @classmethod
    def get_generic_type(cls, type_tag):
        lookup = {
            f'device_type.{tag}': sub_cls.__name__.lower()
            for sub_cls in iter_subclasses(cls)
            for tag in sub_cls.tags
        }
        # These device types are all considered 'occupancy' but could apparently
        # also be multi-sensors based on their state.
        unknowns = (
            'room_sensor',
            'temperature_sensor',
            # LM = LIGHT MOTION?
            'lm',
            'pir',
            'povs',
        )
        lookup.update((f'device_type.{unknown}', 'unknown') for unknown in unknowns)
        return lookup.get(type_tag.lower())



import platformdirs


class PlatformDirs(platformdirs.PlatformDirs):  # type: ignore
    """
    >>> dirs = PlatformDirs(appname='Abode', appauthor=False)
    >>> alt_udp = getfixture('tmp_path') / 'data' / 'dir'
    >>> vars(dirs).update(user_data_path=alt_udp)
    >>> assert dirs.user_data.is_dir()
    """

    @property
    def user_data(self):
        self.user_data_path.mkdir(parents=True, exist_ok=True)
        return self.user_data_path


paths = PlatformDirs(appname='Abode', appauthor=False)



import functools

import jaraco.itertools
import jaraco.functools
import more_itertools

single = jaraco.functools.compose(more_itertools.only, jaraco.itertools.always_iterable)

opt_single = jaraco.functools.compose(
    functools.partial(more_itertools.only, default=None),
    jaraco.itertools.always_iterable,
)



"""Test the Abode binary sensors."""
import functools
import itertools

from jaraco.abode.helpers import urls
import jaraco.abode.devices.status as STATUS

from .mock import login as LOGIN
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import logout as LOGOUT
from .mock import panel as PANEL
from .mock.devices import door_contact as DOOR_CONTACT
from .mock.devices import glass as GLASS
from .mock.devices import keypad as KEYPAD
from .mock.devices import remote_controller as REMOTE_CONTROLLER
from .mock.devices import siren as SIREN
from .mock.devices import status_display as STATUS_DISPLAY
from .mock.devices import water_sensor as WATER_SENSOR


class TestBinarySensors:
    """Test the binary sensors."""

    def test_binary_sensor_properties(self, m):
        """Tests that binary sensor device properties work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))

        # Set up all Binary Sensor Devices in "off states"
        all_devices = [
            DOOR_CONTACT.device(
                devid=DOOR_CONTACT.DEVICE_ID,
                status=STATUS.CLOSED,
                low_battery=False,
                no_response=False,
            ),
            GLASS.device(
                devid=GLASS.DEVICE_ID,
                status=STATUS.OFFLINE,
                low_battery=False,
                no_response=False,
            ),
            KEYPAD.device(
                devid=KEYPAD.DEVICE_ID,
                status=STATUS.OFFLINE,
                low_battery=False,
                no_response=False,
            ),
            REMOTE_CONTROLLER.device(
                devid=REMOTE_CONTROLLER.DEVICE_ID,
                status=STATUS.OFFLINE,
                low_battery=False,
                no_response=False,
            ),
            SIREN.device(
                devid=SIREN.DEVICE_ID,
                status=STATUS.OFFLINE,
                low_battery=False,
                no_response=False,
            ),
            STATUS_DISPLAY.device(
                devid=STATUS_DISPLAY.DEVICE_ID,
                status=STATUS.OFFLINE,
                low_battery=False,
                no_response=False,
            ),
            WATER_SENSOR.device(
                devid=WATER_SENSOR.DEVICE_ID,
                status=STATUS.OFFLINE,
                low_battery=False,
                no_response=False,
            ),
        ]

        m.get(urls.DEVICES, json=all_devices)

        # Logout to reset everything
        self.client.logout()

        # Test our devices
        for device in self.client.get_devices():
            assert not device.is_on, device.type + " is_on failed"
            assert not device.battery_low, device.type + " battery_low failed"
            assert not device.no_response, device.type + " no_response failed"

        # Set up all Binary Sensor Devices in "off states"
        all_devices = [
            DOOR_CONTACT.device(
                devid=DOOR_CONTACT.DEVICE_ID,
                status=STATUS.OPEN,
                low_battery=True,
                no_response=True,
            ),
            GLASS.device(
                devid=GLASS.DEVICE_ID,
                status=STATUS.ONLINE,
                low_battery=True,
                no_response=True,
            ),
            KEYPAD.device(
                devid=KEYPAD.DEVICE_ID,
                status=STATUS.ONLINE,
                low_battery=True,
                no_response=True,
            ),
            REMOTE_CONTROLLER.device(
                devid=REMOTE_CONTROLLER.DEVICE_ID,
                status=STATUS.ONLINE,
                low_battery=True,
                no_response=True,
            ),
            SIREN.device(
                devid=SIREN.DEVICE_ID,
                status=STATUS.ONLINE,
                low_battery=True,
                no_response=True,
            ),
            STATUS_DISPLAY.device(
                devid=STATUS_DISPLAY.DEVICE_ID,
                status=STATUS.ONLINE,
                low_battery=True,
                no_response=True,
            ),
            WATER_SENSOR.device(
                devid=WATER_SENSOR.DEVICE_ID,
                status=STATUS.ONLINE,
                low_battery=True,
                no_response=True,
            ),
        ]

        m.get(urls.DEVICES, json=all_devices)

        # Refesh devices and test changes
        for device in skip_alarms(self.client.get_devices(refresh=True)):
            assert device.is_on, device.type + " is_on failed"
            assert device.battery_low, device.type + " battery_low failed"
            assert device.no_response, device.type + " no_response failed"


def is_alarm(device):
    return device.type_tag == 'device_type.alarm'


skip_alarms = functools.partial(itertools.filterfalse, is_alarm)
